/* 
   $Id: IMRecieve.c,v 1.9 2001/12/25 05:30:39 sho Exp $

   Copyright (C) 2001,2002 AXE,Inc. <sg@axe-inc.co.jp>

   This file is part of Nunome.

   Nunome is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   Nunome is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Nunome; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xproto.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xlocale.h>
#include <X11/keysym.h>

#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Label.h>

#include "IMProto.h"
#include "IMlib.h"
#include "nnmUI.h"

/* IM Connection */
extern int		NumImConnection;
extern ImConnectionRec	*im_connection;

int	NumInputContext = 0;
XIC_Rec	*InputContext=NULL;

static int	CurrentICID = 0;
static int	CurrentIMID = 0;

void SendXIM_ERROR(Window sender, Widget server_w, int imID, int icID,
  int mask, int ecode)
{
    unsigned char pckt[2048];
    int len;

    len = CreatePacketXIM_ERROR(pckt,sender,imID,icID,mask,ecode);
    SendXIMPacket( XtDisplay(server_w), sender, XtWindow(server_w), pckt, len );
}
  
/*
 *	Recieve XIM_CONNECT packet
 */
void	RecieveXIM_CONNECT( unsigned char *packet, int len, 
			    Window sender, Widget server_w )
{
  Display		*dpy = XtDisplay( server_w );
  int			i;
  unsigned char		pckt[64];
  
  if ( sender ){
    if (packet[0] != 0x6c) {
      fprintf(stderr,"Not supported Byte order\n");
      goto sendng;
    }
    if (XimPcktGetCARD16(&(packet[6])) != 0) {
      fprintf(stderr,"Authentication is not supported\n");
      goto sendng;
    }
#ifdef DEBUG_PRINT
    fprintf(stderr," XIM Protocol: Recieve XIM_CONNECT --> Send XIM_CONNECT_REPLY\n");
#endif
    i = 0;
    pckt[i] = XIM_CONNECT_REPLY; i++;	// major-opcode
    pckt[i] = 0; i++;			// minor-opcode
    i += XimPcktSetCARD16( &(pckt[i]), 1 );  // length(4byte unit)
    i += XimPcktSetCARD16( &(pckt[i]), 0x0001 ); // server-protocol-version
    i += XimPcktSetCARD16( &(pckt[i]), 0x0000 ); 
    SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, i );
    return;
sendng:
    i = 0;
    pckt[i] = XIM_AUTH_NG; i++;	// major-opcode
    pckt[i] = 0; i++;			// minor-opcode
    i += XimPcktSetCARD16( &(pckt[i]), 0 );  // length(4byte unit)
    SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, i );
    DisposeIMConnection(sender);
  }
}

/*
 *	Recieve XIM_DISCONNECT packet
 */
void	RecieveXIM_DISCONNECT( unsigned char *packet, int len, 
			       Window sender, Widget server_w )
{
  Display		*dpy = XtDisplay( server_w );
  int			i;
  unsigned char		pckt[64];
  
  if ( sender ){
#ifdef DEBUG_PRINT
    fprintf(stderr," XIM Protocol: Recieve XIM_DISCONNECT --> Send XIM_DISCONNECT_REPLY\n");
#endif
    i = 0;
    pckt[i] = XIM_DISCONNECT_REPLY; i++;	// major-opcode
    pckt[i] = 0; i++;				// minor-opcode
    i += XimPcktSetCARD16( &(pckt[i]), 0 );  // length(4byte unit)
    SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, i );
    DisposeIMConnection(sender);
  }
}


/*
 *	Recieve XIM_OPEN packet
 */
void	RecieveXIM_OPEN( unsigned char *packet, int len, 
			 Window sender, Widget server_w )
{
  Display		*dpy = XtDisplay( server_w );
  unsigned char		pckt[2048];
  int			pckt_len;
  int			r;
  
  // Property with ClientMessageEvent
  if ( sender ){
#ifdef DEBUG_PRINT
    fprintf(stderr," XIM Protocol: Recieve XIM_OPEN --> Send XIM_OPEN_REPLY\n");
#endif
    if ((pckt_len = CreatePacketXIM_REGISTER_TRIGGERKEYS( &(pckt[0]), 
						     sender )) < 0) {
      SendXIM_ERROR(sender,server_w,0,0,0,-pckt_len);
      return;
    }
    if ((r = CreatePacketXIM_OPEN_REPLY( &(pckt[pckt_len]), sender )) < 0) {
      SendXIM_ERROR(sender,server_w,0,0,0,-pckt_len);
      return;
    }
    pckt_len += r;
    SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, pckt_len );
  }
  
}



void	RecieveXIM_QUERY_EXTENSION( unsigned char *packet, int len, 
				    Window sender, Widget server_w )
{
  Display		*dpy = XtDisplay( server_w );
  unsigned char		pckt[2048];
  int			pckt_len, imID;
  
  // Property with ClientMessageEvent
  if ( sender ){
#ifdef DEBUG_PRINT
    fprintf(stderr," XIM Protocol: Recieve XIM_QUERY_EXTENSION -->"
	   "Send XIM_QUERY_EXTENSION_REPLY\n");
#endif // DEBUG_PRINT
    // Property Setting
    XimPcktGetXIM_QUERY_EXTENSION( packet, &imID );
    pckt_len = CreatePacketXIM_QUERY_EXTENSION_REPLY( pckt, sender, imID );
    SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, pckt_len );
  }
}


static char	default_encoding[] = "COMPOUND_TEXT";
//static char	default_encoding[] = "eucJP";

void	RecieveXIM_ENCODING_NEGOTIATION( unsigned char *packet, int len,
					 Window sender, Widget server_w )
{
  Display		*dpy = XtDisplay( server_w );
  unsigned char		pckt[2048];
  int			pckt_len, imID, i, idx;
  int			slen, ilen, length, l;
  XIMSTR		*ximstr;
  XIMENCODINGINFO	*ximencodinginfo;

  if ( sender ){
#ifdef DEBUG_PRINT
    fprintf(stderr," XIM Protocol: Recieve XIM_ENCODING_NEGOTIATION -->"
	   "Send XIM_ENCODING_NEGOTIATION_REPLY\n");
#endif
    length = strlen( default_encoding );
    XimPcktGetXIM_ENCODING_NEGOTIATION( packet, &imID, &ximstr, &slen,
					&ximencodinginfo, &ilen );
    for( i=0; i<slen; i++ ){
      l = ( length < ximstr[i].len )? length:ximstr[i].len;
      if ( strncmp( ximstr[i].ptr, default_encoding,  l )==0 ){
	break;
      }
    }
    free(ximstr);
    free(ximencodinginfo);
    if ( i<slen ){
      idx = i;
    }else{
      fprintf(stderr," Encoding Unmatch\n");
      idx = -1;
      return ;
    }
    pckt_len = CreatePacketXIM_ENCODING_NEGOTIATION_REPLY( pckt, sender,
							   imID, idx );
    SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, pckt_len );
  }
}



/*
 *	Recieve XIM_CLOSE packet
 */
void	RecieveXIM_CLOSE( unsigned char *packet, int len,
			  Window sender, Widget server_w )
{
  Display		*dpy = XtDisplay( server_w );
  unsigned char		pckt[2048];
  int			pckt_len, imID;

  if ( sender ){
#ifdef DEBUG_PRINT
    fprintf(stderr," XIM Protocol: Recieve XIM_CLOSE -->"
	   "Send XIM_CLOSE_REPLY\n");
#endif
    imID = XimPcktGetCARD16( packet );
    
    pckt_len = CreatePacketXIM_CLOSE_REPLY( pckt, sender,
					    imID );
    SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, pckt_len );
  }
}



/*
 *	Recieve XIM_SET_IM_VALUES packet
 */
void	RecieveXIM_SET_IM_VALUES( unsigned char *packet, int len,
				  Window sender, Widget server_w )
{
  Display		*dpy = XtDisplay( server_w );
  unsigned char		pckt[2048];
  int			i, pckt_len, imID;

  if ( sender ){
#ifdef DEBUG_PRINT
    fprintf(stderr," XIM Protocol: Recieve XIM_SET_IM_VALUES -->"
	   "Send XIM_SET_IM_VALUES_REPLT\n");
#endif
    i = 0;
    imID = XimPcktGetCARD16( packet ); i+=2;
#if 0
    {
      XIMATTRIBUTE		ximatr;
      int			imID, idx, n, j, l;
    if ((idx = XimGetIMidx(imID)) < 0) {
      fprintf(stderr," XIM_SET_IM_VALUES: Cannot find imID\n");
      SendXIM_ERROR(sender,server_w,0,0,0,XIMERR_BADSOMETHING);
      return ;
    }
    len  = XimPcktGetCARD16( &(packet[i]) ); i+=2;
    for( n=l=0; l<len; n++ ){
      j = XimPcktGetXIMATTRIBUTE( &(packet[i]), &ximatr );
      XimSetImValue( idx, &ximatr );
      i += j; l += j;
    }
    }
#endif
    pckt_len = CreatePacketXIM_SET_IM_VALUES_REPLY( pckt, sender,
						    imID );
    SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, pckt_len );
  }
}



/*
 *	Recieve XIM_GET_IM_VALUES packet
 */
void	RecieveXIM_GET_IM_VALUES( unsigned char *packet, int len,
				  Window sender, Widget server_w )
{
  Display		*dpy = XtDisplay( server_w );
  unsigned char		pckt[2048];
  int			pckt_len, i, imID, idx, n;
  int			attr_ID[1];

  if ( sender ){
#ifdef DEBUG_PRINT
    fprintf(stderr," XIM Protocol: Recieve XIM_GET_IM_VALUES -->"
	   "Send XIM_GET_IM_VALUES_REPLY\n");
#endif
    i = 0;
    imID = XimPcktGetCARD16( packet ); i+=2;
    if ((idx = XimGetIMidx(imID)) < 0) {
      fprintf(stderr," XIM_GET_IM_VALUES: Cannot find im_connection\n");
      SendXIM_ERROR(sender,server_w,0,0,0,XIMERR_BADSOMETHING);
      return ;
    }
    len  = XimPcktGetCARD16( &(packet[i]) ); i+=2;
    len /= 2;
    if (len > 1) {
      SendXIM_ERROR(sender,server_w,0,0,0,XIMERR_BADSOMETHING);
      return ;
    }
    for( n=0; n<len; n++ ){
      attr_ID[n] = XimPcktGetCARD16( &(packet[i]) );
      i += 2;
    }
    if ((pckt_len = CreatePacketXIM_GET_IM_VALUES_REPLY( pckt, sender,
		    idx, attr_ID, n )) < 0) {
      SendXIM_ERROR(sender,server_w,0,0,0,-pckt_len);
      return;
    }
    SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, pckt_len );
  }
}

/*
 *	Recieve XIM_CREATE_IC packet
 */
void	RecieveXIM_CREATE_IC( unsigned char *packet, int len,
			      Window sender, Widget server_w )
{
  Display		*dpy = XtDisplay( server_w );
  unsigned char		pckt[2048];
  int			pckt_len, i, imID, idx, n, j, l, icID;
  XICATTRIBUTE		xicatr;

  if ( sender ){
#ifdef DEBUG_PRINT
    fprintf(stderr," XIM Protocol: Recieve XIM_CREATE_IC -->"
	   "Send XIM_CREATE_IC_REPLY\n");
#endif
    i = 0;
    imID = XimPcktGetCARD16( packet ); i+=2;
    if ((idx = XimGetIMidx(imID)) < 0) {
      fprintf(stderr," XIM_CREATE_IC: Cannot find im_connection\n");
      SendXIM_ERROR(sender,server_w,0,0,0,XIMERR_BADSOMETHING);
      return ;
    }
    len  = XimPcktGetCARD16( &(packet[i]) ); i+=2;
    icID = XimCreateIC(imID);
    for( n=l=0; l<len; n++ ){
      j = XimPcktGetXICATTRIBUTE( &(packet[i]), &xicatr );
      if (XimSetXICATTRIBUTE( icID, &xicatr ) < 0) {
	XimDestroyIC(imID,icID);
	SendXIM_ERROR(sender,server_w,imID,icID,
	  XIMERR_IMIDMASK|XIMERR_ICIDMASK,XIMERR_BADSOMETHING);
	return;
      }
      i += j; l +=j;
    }
    pckt_len = CreatePacketXIM_CREATE_IC_REPLY( pckt, sender, imID, icID );
    SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, pckt_len );

    // Filter Event Mask --> NoMask
    pckt_len = CreatePacketXIM_SET_EVENT_MASK( pckt, sender, imID, icID,
					       0L, 0L );
    SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, pckt_len );
  }
}

/*
 *	Recieve XIM_DESTROY_IC packet
 */
void	RecieveXIM_DESTROY_IC( unsigned char *packet, int len,
			       Window sender, Widget server_w )
{
  Display		*dpy = XtDisplay( server_w );
  unsigned char		pckt[2048];
  int			pckt_len, i, imID, idx, icID;

  if ( sender ){
#ifdef DEBUG_PRINT
    fprintf(stderr," XIM Protocol: Recieve XIM_DESTROY_IC -->"
	   "Send XIM_DESTROY_IC_REPLY\n");
#endif
    i = 0;
    imID = XimPcktGetCARD16( packet ); i+=2;
    if ((idx = XimGetIMidx(imID)) < 0) {
      fprintf(stderr," XIM_CREATE_IC: Cannot find im_connection\n");
      SendXIM_ERROR(sender,server_w,0,0,0,XIMERR_BADSOMETHING);
      return ;
    }
    icID = XimPcktGetCARD16( &(packet[i]) ); i+=2;
    XimDestroyIC( imID, icID );
    pckt_len = CreatePacketXIM_DESTROY_IC_REPLY( pckt, sender, imID, icID );
    SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, pckt_len );
    
  }
}


/*
 *	Recieve XIM_SET_IC_ATTRIBUTE packet
 */
void	RecieveXIM_SET_IC_VALUES( unsigned char *packet, int len,
				  Window sender, Widget server_w )
{
  Display		*dpy = XtDisplay( server_w );
  unsigned char		pckt[2048];
  int			pckt_len, i, imID, idx, n, j, l, icID;
  XICATTRIBUTE		xicatr;

  if ( sender ){
#ifdef DEBUG_PRINT
    fprintf(stderr," XIM Protocol: Recieve XIM_SET_IC_VALUES -->"
	   "Send XIM_SET_IC_VALUES_REPLY\n");
#endif
    i = 0;
    imID = XimPcktGetCARD16( packet ); i+=2;
    if ((idx = XimGetIMidx(imID)) < 0) {
      fprintf(stderr," XIM_CREATE_IC: Cannot find im_connection\n");
      SendXIM_ERROR(sender,server_w,0,0,0,XIMERR_BADSOMETHING);
      return ;
    }
    icID = XimPcktGetCARD16( &(packet[i]) ); i+=2;
    len  = XimPcktGetCARD16( &(packet[i]) ); i+=2;
    i+=2;	// unused
    for( n=l=0; l<len; n++ ){
      j = XimPcktGetXICATTRIBUTE( &(packet[i]), &xicatr );
      XimSetXICATTRIBUTE( icID, &xicatr );
      i += j; l +=j;
    }
    pckt_len = CreatePacketXIM_SET_IC_VALUES_REPLY( pckt, sender, imID, icID );
    SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, pckt_len );
  }
}


/*
 *	Recieve XIM_GET_IC_ATTRIBUTE packet
 */
void	RecieveXIM_GET_IC_VALUES( unsigned char *packet, int len,
				  Window sender, Widget server_w )
{
  Display		*dpy = XtDisplay( server_w );
  unsigned char		pckt[2048];
  int			pckt_len, i, imID, idx, n, l, icID;
  int			*xicattrID;

  if ( sender ){
#ifdef DEBUG_PRINT
    fprintf(stderr," XIM Protocol: Recieve XIM_GET_IC_VALUES -->"
	   "Send XIM_GET_IC_VALUES_REPLY\n");
#endif
    i = 0;
    imID = XimPcktGetCARD16( packet ); i+=2;
    if ((idx = XimGetIMidx(imID)) < 0) {
      fprintf(stderr," XIM_CREATE_IC: Cannot find im_connection\n");
      SendXIM_ERROR(sender,server_w,0,0,0,XIMERR_BADSOMETHING);
      return ;
    }
    icID = XimPcktGetCARD16( &(packet[i]) ); i+=2;
    len  = XimPcktGetCARD16( &(packet[i]) ); i+=2;
    xicattrID = (int *)malloc( len*sizeof(int)/sizeof(short) );
    for( n=l=0; l<len; n++ ){
      xicattrID[n] = XimPcktGetCARD16( &(packet[i]) );
      i+=2; l+=2;
    }
    if ((pckt_len = CreatePacketXIM_GET_IC_VALUES_REPLY( pckt, sender, 
						    imID, icID, 
						    xicattrID, n )) < 0) {
	SendXIM_ERROR(sender,server_w,imID,icID,
	  XIMERR_IMIDMASK|XIMERR_ICIDMASK,-pckt_len);
	free(xicattrID);
	return;
    }
    free(xicattrID);
    SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, pckt_len );
  }
}

void	RecieveXIM_TRIGGER_NOTIFY( unsigned char *packet, int len,
				   Window sender, Widget server_w )
{
  Display		*dpy = XtDisplay( server_w );
  unsigned char		pckt[2048];
  int			pckt_len, i, imID, icID;

#ifdef DEBUG_PRINT
  fprintf(stderr," Recieve XIM_TRIGGER_NOTIFY -->"
	 " Reply XIM_TRIGGER_NOTIFY_REPLY, and XIM_SET_EVENT_MASK:");
#endif
  i = 0;
  imID = XimPcktGetCARD16( &(packet[i]) ); i+=2;
  icID = XimPcktGetCARD16( &(packet[i]) ); i+=2;

  pckt_len = CreatePacketXIM_TRIGGER_NOTIFY_REPLY( pckt, sender, imID, icID );
  SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, pckt_len );
#if 0
  { int icidx,flag,idx;
  unsigned long		mask;
  flag = XimPcktGetCARD32( &(packet[i]) ); i+=4;
  idx  = XimPcktGetCARD32( &(packet[i]) ); i+=4;
  mask = XimPcktGetCARD32( &(packet[i]) ); i+=4;
  icidx = XimGetICidx( icID );
  if ( InputContext[icidx].triggered ){
    InputContext[icidx].triggered = 0;
    pckt_len = 
      CreatePacketXIM_SET_EVENT_MASK( pckt, sender, imID, icID,
				      0L, 0L );
    NnmPopdown();
#ifdef DEBUG_PRINT
    fprintf(stderr," Popdown\n");
#endif
  }else{
    InputContext[icidx].triggered = 1;
    pckt_len = 
      CreatePacketXIM_SET_EVENT_MASK( pckt, sender, imID, icID,
				      KeyPressMask|StructureNotifyMask,
				      KeyPressMask|StructureNotifyMask);
    NnmAutoPopup();
#ifdef DEBUG_PRINT
    fprintf(stderr," Popup\n");
#endif
  }
  SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, pckt_len );
  }
#endif
}



void	RecieveXIM_FORWARD_EVENT( unsigned char *packet, int len,
				  Window sender, Widget server_w )
{
#if 0
  Display		*dpy = XtDisplay( server_w );
  unsigned char		pckt[2048];
  int			pckt_len, i, imID, flag, icID;
  int			serial;
  xEvent		sEv;
  XEvent		ev;

  i = 0;
  imID = XimPcktGetCARD16( &(packet[i]) ); i+=2;
  icID = XimPcktGetCARD16( &(packet[i]) ); i+=2;
  flag = XimPcktGetCARD16( &(packet[i]) ); i+=2;
  serial  = XimPcktGetCARD16( &(packet[i]) ); i+=2;
  memcpy( &sEv, &(packet[i]), sizeof(sEv) );
  /* Common part */
  ev.type = sEv.u.u.type & 0x7f;
  ev.xany.serial = ( serial << 16 )| sEv.u.u.sequenceNumber;
  ev.xany.display= dpy;
  ev.xany.send_event = (sEv.u.u.type > 0x7f)? 1:0;
  if ( ev.type != KeyPress ){
    fprintf(stderr,"\tUnsupport Event Type %3d: \n", ev.type );
  }

  ev.xkey.window    = sEv.u.keyButtonPointer.event;
  ev.xkey.root      = sEv.u.keyButtonPointer.root;
  ev.xkey.subwindow = sEv.u.keyButtonPointer.child;
  ev.xkey.time      = sEv.u.keyButtonPointer.time;
  ev.xkey.x         = cvtINT16toInt(sEv.u.keyButtonPointer.eventX);
  ev.xkey.y         = cvtINT16toInt(sEv.u.keyButtonPointer.eventY);
  ev.xkey.x_root    = cvtINT16toInt(sEv.u.keyButtonPointer.rootX);
  ev.xkey.y_root    = cvtINT16toInt(sEv.u.keyButtonPointer.rootY);
  ev.xkey.state     = sEv.u.keyButtonPointer.state;
  ev.xkey.keycode   = sEv.u.u.detail;
  ev.xkey.same_screen = sEv.u.keyButtonPointer.sameScreen;

#ifdef DEBUG_PRINT
  fprintf(stderr," Recieve XIM_FORWARD_EVENT(keycode %3X:state %3X) %X, %X --> ",
	 ev.xkey.keycode, ev.xkey.state,
	 ev.xkey.window, ev.xkey.subwindow );
#endif
  // Return
  if ( ev.xkey.keycode == 0x26 ){
    unsigned char	commitstr[]="";
    XTextProperty	ct;

    XimConvertToCompoundText( dpy, XCompoundTextStyle, commitstr, &ct );

    pckt_len = CreatePacketXIM_COMMIT( pckt, sender, imID, icID,
				       0x0002, ct.value,
				       ct.nitems );
    SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, pckt_len );
#ifdef DEBUG_PRINT
    fprintf(stderr," Reply XIM_COMMIT");
#endif
  }else{
    /*
    pckt_len = CreatePacketXIM_FORWARD_EVENT( pckt, sender, imID, icID,
					      0x0000, serial, 
					      &(packet[8]), len*4-8 );
    */
    pckt_len = CreatePacketXIM_FORWARD_EVENT( pckt, sender, imID, icID,
					      0x0000, serial, &ev );
    SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, pckt_len );
#ifdef DEBUG_PRINT
    fprintf(stderr," Reply XIM_FORWARD_EVENT");
#endif
  }
  if ( flag & 0x0001 ){		// Synchronous bit
    pckt_len = CreatePacketXIM_SYNC_REPLY( pckt, sender, imID, icID );
    SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, pckt_len );
#ifdef DEBUG_PRINT
    fprintf(stderr,",  Reply XIM_SYNC_REPLY");
#endif
  }
#ifdef DEBUG_PRINT
  fprintf(stderr,"\n");
#endif
#else
  int			i,imID,icID;

  i = 0;
  imID = XimPcktGetCARD16( &(packet[i]) ); i+=2;
  icID = XimPcktGetCARD16( &(packet[i]) ); i+=2;
  SendXIM_ERROR(sender,server_w,0,0,0,XIMERR_BADPROTOCOL);
#endif
}



void	RecieveXIM_SYNC( unsigned char *packet, int len,
			 Window sender, Widget server_w )
{
  Display		*dpy = XtDisplay( server_w );
  unsigned char		pckt[2048];
  int			pckt_len, i, imID, icID;

#ifdef DEBUG_PRINT
  fprintf(stderr," Recieve XIM_SYNC --> Reply XIM_SYNC_REPLY\n");
#endif
  i = 0;
  imID = XimPcktGetCARD16( &(packet[i]) ); i+=2;
  icID = XimPcktGetCARD16( &(packet[i]) ); i+=2;

  pckt_len = CreatePacketXIM_SYNC_REPLY( pckt, sender, imID, icID );
  SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, pckt_len );

}


void	RecieveXIM_SET_IC_FOCUS( unsigned char *packet, int len,
				 Window sender, Widget server_w )
{
  int			i, imID, icID;

#ifdef DEBUG_PRINT
  fprintf(stderr," Recieve XIM_SET_IC_FOCUS  --> Popup\n");
#endif
  i = 0;
  imID = XimPcktGetCARD16( &(packet[i]) ); i+=2;
  icID = XimPcktGetCARD16( &(packet[i]) ); i+=2;

  NnmAutoPopup();

  CurrentIMID = imID;
  CurrentICID = icID;
}



void	RecieveXIM_UNSET_IC_FOCUS( unsigned char *packet, int len,
				   Window sender, Widget server_w )
{
#ifdef DEBUG_PRINT
  fprintf(stderr," Recieve XIM_UNSET_IC_FOCUS\n");
#endif
}

void	XimSendString( char *str )
{
  Display		*dpy;
  unsigned char		pckt[2048];
  int			pckt_len, imID, icID, idx;
  XTextProperty		ct;
  Window		sender;
  Widget		server_w;

#ifdef DEBUG_PRINT
  fprintf(stderr," Send String --> " );
#endif
  imID = CurrentIMID;
  icID = CurrentICID;
  idx = XimGetIMidx( imID );
  if ( idx == -1 ){
#ifdef DEBUG_PRINT
    fprintf(stderr,"No Sending \n");
#endif
    return ;
  }
  sender = im_connection[idx].client;

  server_w = im_connection[idx].server;
  dpy = XtDisplay( server_w );
  if (XimConvertToCompoundText( dpy, XCompoundTextStyle, str, &ct )
    != Success) {
      return;
  }
  pckt_len = CreatePacketXIM_COMMIT( pckt, sender, imID, icID,
				     0x0002, ct.value,
				     ct.nitems );
  SendXIMPacket( dpy, sender, XtWindow(server_w), pckt, pckt_len );
  XFree(ct.value);
#ifdef DEBUG_PRINT
  fprintf(stderr," Sending XIM_COMMIT\n");
#endif
}

void	XimSendKey(unsigned int code)
{
  char str[256];

  if((code >> 8)){
    str[0] = (code >> 8)   | 0x80;
    str[1] = (code & 0xff) | 0x80;
    str[2] = 0;
  }else{
    str[0] = code & 0xff;
    str[1] = 0;
  }

  XimSendString(str);
}
