/* 
   $Id: nnmWindow.c,v 1.10 2002/06/27 05:51:26 shibajun Exp $

   Copyright (C) 2001,2002 AXE,Inc. <sg@axe-inc.co.jp>

   This file is part of Nunome.

   Nunome is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   Nunome is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Nunome; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "config.h"
#include <stdio.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#ifdef __NUNOME_VGA__
#include <X11/Xaw/Box.h>
#endif

#include "./nnmUI.h"

#include "NnmProto.h"
#include "IMlib.h"

#define ATOM_AXE_SG_NUNOME	"_AXE_SG_NUNOME"

static Widget	NnmTopShell;
static Widget	NnmTopForm;

#define KEYBOARD
#ifdef KEYBOARD
static Widget	NnmKBDForm;
static Widget	NnmKigouForm;
#endif /* KEYBOARD */

/*
 *  NnmμPopup WindowFocus뤫ɤ
 *
 *  0: focusʤ;  1: focus
 */
int	NnmIsFocus(void)
{
  int		dummy;
  Window	focus_win;
  Display	*dpy;

  dpy = XtDisplay( NnmTopShell );
  XGetInputFocus( dpy, &focus_win, &dummy );
#ifdef DEBUG_PRINT
  fprintf(stderr,"Focus %X\n",focus_win );
#endif
  if ( XtWindow(NnmTopShell)==focus_win ) return 1;
  if ( XtWindow(NnmTopForm)==focus_win ) return 1;
  return 0;
}


/*
 *  NnmμWindowPopup
 *
 *  ( x, y )Popup֤ꤹ롣
 *
 */
void	NnmPopup()
{
  XtPopup( NnmTopShell, XtGrabNonexclusive );
}

void	NnmAutoPopup()
{
  extern int IsPopup;

  if (IsPopup) NnmPopup();
}


/*
 *  NnmμWindowPopdown
 *
 */
void	NnmPopdown( void )
{
  XtPopdown( NnmTopShell );
  NnmCandPopdown();
}


/*
 *  ǧ̤ɽ뤿FormWidgetκ
 */
static Widget	CreateResultWidgets( char *name, Widget parent, 
				     Arg args[], int num )
{
  Widget	ResultForm;

#ifdef __NUNOME_VGA__
  ResultForm = XtCreateManagedWidget( name, boxWidgetClass,
				      parent, args, num );
#else
  ResultForm = XtCreateManagedWidget( name, formWidgetClass,
				      parent, args, num );
#endif
  NnmCreateResultWidgets( ResultForm );
  return ResultForm;
}


/*
 *  Canvas WindowѤFormWidgetκ
 *
 */
static Widget	CreateCanvasWidgets( char *name, Widget parent,
				     Arg args[], int num )
{
  Widget	CanvasForm;
#ifdef __NUNOME_VGA__
  Widget	DummyBox ;
#endif

#ifdef __NUNOME_VGA__
  int i = num ;
  XtSetArg( args[i], XtNvSpace, 0 ); i++;
  XtSetArg( args[i], XtNhSpace, 0 ); i++;

  DummyBox = XtCreateManagedWidget( name, boxWidgetClass,
                                      parent, args, i );

  CanvasForm = XtCreateManagedWidget( name, formWidgetClass,
                                      DummyBox, args, num );
#else
  CanvasForm = XtCreateManagedWidget( name, formWidgetClass,
				      parent, args, num );
#endif
  NnmCreateCanvases( CanvasForm );
  return CanvasForm;
}


/*
 *
 */
Boolean /*XtEventHandler*/ NnmForcePopdown( Widget w, XtPointer clo, 
				 XEvent *ev, Boolean *ctd )
{
  if ( ev->type == UnmapNotify ){
    NnmPopdown();
    return True;
  }
  return False;
}


/*
 *  Popup Windowκ
 *
 */

Widget	CanvasForm, CommandForm, ResultForm;

void NnmGetShellPos(x, y, w, h)
     int  *x, *y, *w, *h;
{
  NnmGetWindowPos(NnmTopShell, x, y, w, h);
}

void NnmGetResultFormPos(x, y, w, h)
     int  *x, *y, *w, *h;
{
  NnmGetWindowPos(ResultForm, x, y, w, h);
}

void	NnmCreatePopShell( Widget top )
{
  Arg		args[10];
  int		i;
  int		data = 0;
  Atom		NnmAtom_SG_NUNOME;
  Dimension width,height;

  NnmAtom_SG_NUNOME = XInternAtom( XtDisplay(top), ATOM_AXE_SG_NUNOME,
				   False );

  i = 0;
  XtSetArg( args[i], XtNinput, False ), i++;
  NnmTopShell = XtCreatePopupShell( "nnmtop",
				    transientShellWidgetClass,
				    top, args, i );

  NnmTopForm = XtCreateManagedWidget( "topform", formWidgetClass,
				      NnmTopShell, NULL, 0 );
#ifdef KEYBOARD
  NnmKBDForm = XtCreateManagedWidget( "kbdtopform", formWidgetClass,
				      NnmTopShell, NULL, 0 );

  NnmKigouForm = XtCreateManagedWidget( "kigoutopform", formWidgetClass,
					NnmTopShell, NULL, 0 );
#endif /* KEYBOARD */ 

  XtAddEventHandler( NnmTopShell, StructureNotifyMask, False, 
		     (XtEventHandler)NnmForcePopdown, NULL );

  /* Display Results Form Widget */
  i = 0;
  XtSetArg( args[i], XtNfromVert,  NULL ); i++;
  XtSetArg( args[i], XtNfromHoriz, NULL ); i++;
  XtSetArg( args[i], XtNborderWidth, 0 ); i++;
#ifdef __NUNOME_VGA__
  XtSetArg( args[i], XtNorientation, XtorientHorizontal); i++;
  XtSetArg( args[i], XtNvSpace, 0 ); i++;
  XtSetArg( args[i], XtNhSpace, 0 ); i++;
#endif
  ResultForm = CreateResultWidgets( "resultform", NnmTopForm, args, i );

  /* From Widget for Canvas Widgets */
  i = 0;
  XtSetArg( args[i], XtNfromVert,  ResultForm ); i++;
  XtSetArg( args[i], XtNfromHoriz, NULL ); i++;
  XtSetArg( args[i], XtNborderWidth, 0 ); i++;
  CanvasForm = CreateCanvasWidgets( "canvasform", NnmTopForm, args, i );
  
  /* From Widget for Command Widgets */
  i = 0;
  XtSetArg( args[i], XtNborderWidth, 0 ); i++;
  CommandForm = CreateCommandWidgets( "commandform",
				      NnmTopForm, args, i );

  NnmCreateCandidatePopup( top );

#ifdef WNN
  NnmCreateKouhoPopup(  top );
#endif /* WNN */

  NnmCreateJISPopup(  top );

#ifdef KEYBOARD
  i = 0;
  XtSetArg( args[i], XtNborderWidth, 0 ); i++;
  CreateKBDWidgets( "kbdform",
 		    NnmKBDForm, args, i );

  XtSetArg( args[i], XtNborderWidth, 0 ); i++;
  CreateKigouWidgets( "kigouform",
		      NnmKigouForm, args, i );
#endif /* KEYBOARD */

  XtRealizeWidget( NnmTopShell );
  i = 0;
  XtSetArg( args[i], XtNheight , &height ); i++;
  XtSetArg( args[i], XtNwidth, &width ); i++;
  XtGetValues(NnmTopShell,args,i);
  i = 0;
  XtSetArg( args[i], XtNmaxHeight , height ); i++;
  XtSetArg( args[i], XtNmaxWidth, width ); i++;
  /*
  XtSetArg( args[i], XtNminHeight , height ); i++;
  XtSetArg( args[i], XtNminWidth, width ); i++;
  */
  XtSetValues(NnmTopShell,args,i);

  XChangeProperty( XtDisplay(NnmTopShell), XtWindow(NnmTopShell),
		   NnmAtom_SG_NUNOME, XA_CARDINAL, 32, PropModeAppend,
		   (unsigned char *)&data, 1 );
}

void UnmapTopForm()
{
  XtUnmapWidget(NnmTopForm);
}

void MapTopForm()
{
  XtMapWidget(NnmTopForm);
}

#ifdef KEYBOARD
void UnmapKBDForm()
{
  XtUnmapWidget(NnmKBDForm);
}

void MapKBDForm()
{
  XtMapWidget(NnmKBDForm);
}

void UnmapKigouForm()
{
  XtUnmapWidget(NnmKigouForm);
}

void MapKigouForm()
{
  XtMapWidget(NnmKigouForm);
}
#endif /* KEYBOARD */
