
/* 
   $Id: nnmime.c,v 1.15 2002/06/27 05:51:26 shibajun Exp $

   Copyright (C) 2001,2002 AXE,Inc. <sg@axe-inc.co.jp>

   This file is part of Nunome.

   Nunome is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   Nunome is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Nunome; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "config.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <locale.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xproto.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xlocale.h>

#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Label.h>

#include "./nnmUI.h"

#include "NnmProto.h"
#include "IMlib.h"

#define KEYBOARD

#include "nnmclass.h"
#include "nnmStub.h"
#include "nnmCanvas.h"

static XtResource resources[] = {
  { XtNcallback, XtCCallback, XtRCallback, sizeof(XtPointer),
    XtOffset(NunomeWidget, nunome.callback), XtRCallback, (XtPointer)NULL},
  { "jserver", "jserver", XtRString, sizeof(String),
    XtOffset(NunomeWidget, nunome.jserver), XtRString, NULL},
  { "wnnenvrc", "wnnenvrc", XtRString, sizeof(String),
    XtOffset(NunomeWidget, nunome.wnnenvrc), XtRString, 
    NULL},
  { "dictionary", "dictionary", XtRString, sizeof(String),
    XtOffset(NunomeWidget, nunome.dictionary), XtRString, 
    NULL},
  { "userdictionary", "userdictionary", XtRString, sizeof(String),
    XtOffset(NunomeWidget, nunome.user_dictionary), XtRString, 
    NULL},
#ifdef KEYBOARD
  { "keymap", "keymap", XtRString, sizeof(int),
    XtOffset(NunomeWidget, nunome.keymap), XtRString, NULL},
#endif
  { "nolearn", "nolearn", XtRBoolean, sizeof(Boolean),
    XtOffset(NunomeWidget, nunome.nolearn), XtRBoolean, False},
  { "popup", "popup", XtRBoolean, sizeof(Boolean),
    XtOffset(NunomeWidget, nunome.popup), XtRBoolean, False},
  { "writeSample", "writeSample", XtRBoolean, sizeof(Boolean),
    XtOffset(NunomeWidget, nunome.writeSample), XtRBoolean, False},
  { "timeout", "timeout", XtRInt, sizeof(int),
    XtOffset(NunomeWidget, nunome.timeout), XtRInt, NULL},
};

int recogTimeout;
int serverPID;

static void realize(){}

NunomeClassRec nunomeClassRec = {
  {
    &widgetClassRec,
    "Nunome",
    sizeof(NunomeRec),
    NULL,
    NULL,
    False,
    NULL,/* init */
    NULL,
    realize,/* realize */
    NULL,
    0,
    resources,
    XtNumber(resources),
    (XrmClass) NULL,
    True,
    True,
    True,
    False,
    NULL,/* destroy */
    NULL,
    NULL, /* redisplay */
    NULL, /* set_values */
    NULL, /* set_values_hook */
    NULL, /* set_valuse_almost */
    NULL,
    NULL,
    XtVersion,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL
  }
};

WidgetClass nunomeWidgetClass = (WidgetClass) &nunomeClassRec;

/* Callback Routine on Button for Quit */
static void	Quit( Widget w, XtPointer cli, XtPointer clo )
{
  exit(0);
}


/* 
 *  XError Handler
 *
 */
static int	privErrHandler( Display *dpy, XErrorEvent *ErrEv )
{
  char	string[2048];

  XGetErrorText( dpy, ErrEv->error_code, string, 2048 );
  printf("ErroMsg: %s\n", string );
  printf(" ErrorCode %d: Request %d: Minor %d\n",
	 ErrEv->error_code, ErrEv->request_code, ErrEv->minor_code );
  return 0;
}

int  IsPopup       = 0;
int  IsWriteSample = 0;
static int setPopup = 0;
static int setWriteSample = 0;
static char *dictionary = NULL;
char *user_dictionary = NULL;
int learn_flag = 1;
#ifdef KEYBOARD
char *nnm_keymap = NULL;
#endif /* KEYBOARD */

void usage(char *cmd)
{
    fprintf(stderr,"Usage:%s [+popup/-popup][--write-sample]\n"
      "\t[-dictionary <dictionary file name>]\n"
#ifdef KEYBOARD
      "\t[-nnmkeymap <keymap name>]\n"
#endif
      "\t[-userdictionary <user dictionary file name>][-nolearn]\n",
      cmd);
    exit(2);
}

void NnmParseArguments(int argc, char **argv)
{
  int i;

  for (i = 1;i < argc;i++) {
    if(strcmp( "+popup", argv[i]) == 0){
      IsPopup = 1;
      setPopup = 1;
    }else if(strcmp( "-popup", argv[i]) == 0){
      IsPopup = 0;
      setPopup = 1;
    }else if(strcmp( "--write-sample", argv[i]) == 0){
      IsWriteSample = 1;
      setWriteSample = 1;
    }else if(strcmp( "-dictionary", argv[i]) == 0){
	if (++i >= argc) {
	  usage(argv[0]);
	}
	dictionary = argv[i];
    }else if(strcmp( "-userdictionary", argv[i]) == 0){
	if (++i >= argc) {
	  usage(argv[0]);
	}
	user_dictionary = argv[i];
#ifdef KEYBOARD
    }else if(strcmp( "-nnmkeymap", argv[i]) == 0){
      if (++i >= argc) {
	usage(argv[0]);
      }
      nnm_keymap = argv[i];
#endif /* KEYBOARD */
    }else if(strcmp( "-nolearn", argv[i]) == 0){
	learn_flag = 0;
    } else {
      usage(argv[0]);
    }
 }
}

void NnmGetResource(w, name, value_return)
     Widget w;
     String name;
     XtArgVal *value_return;
{
  Arg arg;

  arg.name = name;
  arg.value = (XtArgVal) value_return; 
  XtGetValues( w, &arg, 1);
}

static int server_ifd;

static void startServer()
{
    int infd[2];
    int outfd[2];
    FILE *ifp,*ofp;


    if (pipe(infd) < 0) {
	fprintf(stderr,"pipe fail\n");
	exit(2);
    }
    if (pipe(outfd) < 0) {
	fprintf(stderr,"pipe fail\n");
	exit(2);
    }
    if ((serverPID = fork()) == 0) {
	close(infd[0]);
	close(outfd[1]);
	dup2(infd[1],1);
	dup2(outfd[0],0);
	execlp(NNMSERVER,NNMSERVER,0);
	fprintf(stderr,"Exec server fail\n");
	exit(2);
    }
    close(infd[1]);
    close(outfd[0]);
    if ((ifp = fdopen(infd[0],"r")) == NULL) {
	fprintf(stderr,"fdopen fail\n");
	exit(2);
    }
    if ((ofp = fdopen(outfd[1],"w")) == NULL) {
	fprintf(stderr,"fdopen fail\n");
	exit(2);
    }
    NnmStubInit(ofp,ifp);
    server_ifd = infd[0];
}

static void handleServerEvent(XtPointer client_data, int *src, XtInputId *id)
{
    NnmHandleRecognizeResult();
}


/*
 *  NnmMain
 *
 *    appShellWidgetmapʤ̿ѡ
 *    ȤΤWindowPopupShell
 */
int	main( int argc, char **argv )
{
  XtAppContext	app_con;
  Widget	toplevel, child;
  char		*support_locale;
  int           num;
  Arg		args[10];
  struct stat	st;

  Boolean popup,wr_sam;

  /* Locale */  
  support_locale = setlocale( LC_ALL, "" );
#ifdef DEBUG_PRINT
  printf("%s\n",support_locale );
#endif
  XtSetLanguageProc( NULL, NULL, NULL );	// Xtlocaleؿ

  /* XErrorΥϥɥ */
  XSetErrorHandler( privErrHandler );

  /* application top shell widget */
  num = 0;
  XtSetArg( args[num], XtNwidth, 1 ),num++;
  XtSetArg( args[num], XtNheight, 1 ),num++;
  XtSetArg( args[num], XtNmappedWhenManaged, False ),num++;
  toplevel = XtAppInitialize( &app_con, "Nunome", NULL, 0,
			      &argc, argv, NULL, args, num );

  NnmParseArguments(argc, argv);

  /*
   *  dummychild widget
   *    toplevel widgetsize0errorˤʤ뤿child widget
   */
  child = XtCreateManagedWidget( "label", nunomeWidgetClass,
				 toplevel, args, 0 );
  XtAddCallback( child, XtNcallback, Quit, NULL );

  if(!setPopup){
    NnmGetResource(child, (String)"popup", (XtArgVal *)&popup);
#ifdef DEBUG_PRINT
    printf("popup: %s\n", popup ? "True" : "False");
#endif
    IsPopup = popup;
  }

  if(!setWriteSample){
    NnmGetResource(child, (String)"writeSample", (XtArgVal *)&wr_sam);
#ifdef DEBUG_PRINT
    printf("write sample: %s\n", wr_sam ? "True" : "False");
#endif
    IsWriteSample = wr_sam;
  }

  if (dictionary == NULL) {
    NnmGetResource(child, (String)"dictionary",(XtArgVal *) &dictionary);
  }
  if (dictionary == NULL) {
#ifndef PKGDATADIR
#define PKGDATADIR "/usr/local/share/nunome"
#endif
    dictionary = PKGDATADIR "/nunome.dic";
  }
  if (user_dictionary == NULL) {
    NnmGetResource(child, (String)"userdictionary",(XtArgVal *) &user_dictionary);
  }
  if (user_dictionary == NULL) {
    char buf[4096];

    char *p = getenv("HOME");
    if (p == NULL) {
	fprintf(stderr,"Can't get value of environment varibale HOME\n");
	exit(2);
    }
    sprintf(buf,"%s/nnmuser.dic",p);
    user_dictionary = strdup(buf);
  }
  if (learn_flag){
    Boolean nolearn;
    NnmGetResource(child, (String)"nolearn", (XtArgVal *)&nolearn);
    if(nolearn) learn_flag = 0;
  }

  NnmGetResource(child, (String)"timeout", (XtArgVal *)&recogTimeout);
  if(recogTimeout <= 0){
    recogTimeout = 1000;
  }

#ifdef KEYBOARD
  if(!nnm_keymap){
    NnmGetResource(child, (String)"keymap", (XtArgVal *)&nnm_keymap);
#ifdef __NUNOME_PDA__
    if(!nnm_keymap) nnm_keymap = "key101";
#elif defined __NUNOME_VGA__
    if(!nnm_keymap) nnm_keymap = "key101vga";
#else
    if(!nnm_keymap) nnm_keymap = "keymg2";
#endif
  }
#endif

#ifdef WNN
  NnmOpenJserver( child );
#endif

  /* NuNoMe Handwritten Window (PopupShell) */
  NnmCreatePopShell( toplevel );
  XtRealizeWidget( toplevel );

  /*
   * Create Atom for Property
   *   ץѥƥtoplevel widgetrealizeƤǤʤ
   * toplevel widget(window)Žդʤ
   */
  if (SetServerProperty( toplevel ) < 0) {
    exit(2);
  }

  /* Set Preconnection Event Handler */
  XtAddEventHandler( toplevel, 0L, True, 
		     (XtEventHandler)TopLevelEvHandler , 
		     (XtPointer)NULL );
  InitSendEventKey(XtDisplay(toplevel));
  NnmAutoPopup();

  startServer();
  NnmStubSetDictionaryFile(dictionary,0);
  if (stat(user_dictionary,&st) < 0) {
    if (creat(user_dictionary,0666) < 0) {
	fprintf(stderr,"Can't create User dictionary:%s\n",user_dictionary);
	exit(2);
    }
  }
  NnmStubSetDictionaryFile(user_dictionary,1);
  XtAppAddInput(app_con,server_ifd,(XtPointer)XtInputReadMask,
    handleServerEvent, NULL);
  NnmInitRecognizeField();
  XtAppMainLoop( app_con );
  return 0;
}

/*
 *	Preconnectionγ
 *
 * (1) XIM Server @im=_NUNOME Ȥ Atom 
 *    ˡXIM_SERVERS Ȥ Atom  Property  @im=_NUNOME Ȥ
 *    Atom ɲä롣
 *     @im=_NUNOME Ȥ Atom Υ쥯Υʡˤʤ롣
 *    ( ؿSetServerProperty() )
 *
 * (2) XIM Client  @im=_NUNOME Ȥ쥯Ф
 *    LOCALES  TRANSPORT Υץѥƥ׵᤹Τ
 *    (SelectionRequest Event)ץѥƥ򥻥åȤơ
 *    XSendEventǡSelectionNotify Event 롣
 *
 * (3) TRANSPORT ץѥƥ X Protocol Ѥ硢
 *    _XIM_CONNECT  ClientMessage Event Ƥ롣
 *    λˡClient window-ID äƤ롣 Client
 *    window ̿뤿 Server window 1-to-1
 *    ̿Ԥ褦ˤ롣XIM ProtocolǤϡClient window
 *     Packet ˴ޤޤʤᡢ̿ꤹ뤿
 *    Server ¦ǤϡClient window 11Ĥб window 
 *    Createɬפ롣
 */
